import pandas as pd
import pandas.io.formats.format as fmt
import zarr


_base_repr = pd.Series.__repr__


def _repr_series(series: pd.Series) -> str:
    if isinstance(series.values, zarr.Array):
        params = fmt.get_series_repr_params()  # pyright: ignore[reportAttributeAccessIssue, reportUnknownVariableType]
        return series.copy().to_string(**params)  # pyright: ignore[reportUnknownVariableType, reportUnknownArgumentType]

    return _base_repr(series)


pd.Series.__repr__ = _repr_series  # pyright: ignore[reportAttributeAccessIssue]



from __future__ import annotations

from collections.abc import Collection, Iterable, Iterator, MutableSequence
from typing import Any, Self, cast, overload, override
import warnings

import numpy as np
import numpy.typing as npt
from numpy._typing import _SupportsArray as SupportsArray  # pyright: ignore[reportPrivateUsage]
import zarr
from zarr.core.common import AccessModeLiteral
from zarr.errors import BoundsCheckError, UnstableSpecificationWarning
from zarr.storage import StoreLike  # pyright: ignore[reportUnknownVariableType]

from ezarr._repr import repr_element
from ezarr.names import UNKNOWN, Attribute, EZType
from ezarr.object import EZObject
from ezarr.types import DeferredCreationFunc, ListWrapper, PyObject, PyObjectCodec


def deferred_EZList(values: Collection[Any]) -> DeferredCreationFunc:
    def inner(grp: zarr.Group, name: str, overwrite: bool) -> None:
        with warnings.catch_warnings(action="ignore", category=UnstableSpecificationWarning):
            arr = grp.create_array(
                name,
                shape=(len(values),),
                dtype=PyObject(),
                serializer=PyObjectCodec(),
                attributes={Attribute.EZType: EZType.List},
                overwrite=overwrite,
            )
        arr[:] = values  # pyright: ignore[reportArgumentType]

    return inner


class EZList[T](EZObject[list[T]], MutableSequence[T]):
    """
    List-like object built on top of a zarr.Array for storing arbitrary Python objects.

    Args:
        array: a zarr.Array with PyObject() dtype

    Example:
        >>> EZList(zarr.array([1, 2, 3]))
        EZList[1, 2, 3]
    """

    def __init__(self, array: zarr.Array):
        assert array.ndim == 1
        self._array: zarr.Array = array

        self._iter_index: int = 0

    @classmethod
    def from_list(
        cls,
        lst: Collection[T],
        *,
        name: str,
        store: StoreLike | None = None,
        mode: AccessModeLiteral = "a",
        path: str | None = None,
        overwrite: bool = False,
    ) -> Self:
        """
        Create an EZList from a list-like object.

        Args:
            lst: list-like collection of values to store in an EZList.
            name: name for the EZList, to use inside the store.
            store: Store, path to a directory or name of a zip file.
            mode: Persistence mode: 'r' means read only (must exist); 'r+' means
                read/write (must exist); 'a' means read/write (create if doesn't
                exist); 'w' means create (overwrite if exists); 'w-' means create
                (fail if exists).
            path: path within the store where the EZList will be saved.
            overwrite: overwrite object if a group with name `name` already exists ? (default: False)

        Example:
            >>> from pathlib import Path
            >>> data = [1, "some text", Path("/some/path.txt")]
            >>> ez_list = EZList.from_list(data, name="data")
            >>> ez_list
            EZList[1, 'some text', PosixPath('/some/path.txt')]

            >>> ez_list = EZList.from_list([1, [2, 3]], name="data")
            >>> ez_list
            EZList[1, array([2, 3])]
        """
        if store is None:
            store = {}

        with warnings.catch_warnings(action="ignore", category=UnstableSpecificationWarning):
            arr = zarr.open_group(store, mode=mode, path=path).create_array(
                name,
                shape=(len(lst),),
                dtype=PyObject(),
                serializer=PyObjectCodec(),
                attributes={Attribute.EZType: EZType.List},
                overwrite=overwrite,
            )
        arr[:] = np.asarray([ListWrapper(e) if isinstance(e, list | SupportsArray) else e for e in lst])  # pyright: ignore[reportUnknownArgumentType]

        return cls(arr)

    def __array__(self, dtype: np.dtype | None = None, copy: bool | None = None) -> npt.NDAarray:
        return self._array[:]

    @override
    def __repr__(self) -> str:
        if len(self) > 6:
            return f"{type(self).__name__}[{repr_element(self[0])}, {repr_element(self[1])}, {repr_element(self[2])}, ..., {repr_element(self[-3])}, {repr_element(self[-2])}, {repr_element(self[-1])}]"

        return f"{type(self).__name__}[{', '.join(repr_element(e) for e in self)}]"

    @override
    def __len__(self) -> int:
        return self._array.shape[0]

    @overload
    def __getitem__(self, item: int) -> T: ...
    @overload
    def __getitem__(self, item: slice[int] | slice[int, int] | slice[int, int, int]) -> Self: ...
    @override
    def __getitem__(self, item: int | slice[int, int, int]) -> T | Self:
        if isinstance(item, slice):
            return type(self)(self._array[item])  # pyright: ignore[reportArgumentType]

        try:
            return cast(T, self._array.get_basic_selection(item)[()])  # pyright: ignore[reportIndexIssue, reportArgumentType, reportCallIssue]

        except BoundsCheckError:
            raise IndexError(
                f"EZList index {item} is out of range [{self._array.shape[0]} ... {self._array.shape[0] - 1}]."
            )

    @overload
    def __setitem__(self, key: int, value: T) -> None: ...
    @overload
    def __setitem__(self, key: slice[int] | slice[int, int] | slice[int, int, int], value: Iterable[T]) -> None: ...
    @override
    def __setitem__(self, key: int | slice[int, int, int], value: T | Iterable[T]) -> None:
        try:
            self._array[key] = np.array(value)

        except BoundsCheckError:
            raise IndexError(
                f"EZList index {key} is out of range [{self._array.shape[0]} ... {self._array.shape[0] - 1}]."
            )

    @override
    def __delitem__(self, key: int | slice[int, int, int]) -> None:
        if isinstance(key, slice):
            raise NotImplementedError

        if key < 0:
            key = len(self) + key

        self._array[key:-1] = self._array[key + 1 :]

        with warnings.catch_warnings(action="ignore", category=UnstableSpecificationWarning):
            self._array.resize(self._array.shape[0] - 1)

    @override
    def __iter__(self) -> Iterator[T]:
        self._iter_index = 0
        return self

    def __next__(self) -> T:
        if self._iter_index >= self._array.shape[0]:
            raise StopIteration

        obj = cast(T, self._array.get_basic_selection(self._iter_index)[()])  # pyright: ignore[reportIndexIssue, reportArgumentType, reportCallIssue]
        self._iter_index += 1

        return obj

    @override
    def __eq__(self, other: object) -> bool:
        if not isinstance(other, list | SupportsArray):
            return False

        return np.array_equal(self._array[()], other)  # pyright: ignore[reportUnknownArgumentType]

    @classmethod
    @override
    def open(
        cls, store: StoreLike | None = None, *, name: str, mode: AccessModeLiteral = "a", path: str | None = None
    ) -> Self:
        """
        Open this EZList from a store.

        Args:
            store: Store, path to a directory or name of a zip file.
            name: name for the EZList, to use inside the store.
            mode: Persistence mode: 'r' means read only (must exist); 'r+' means
                read/write (must exist); 'a' means read/write (create if doesn't
                exist); 'w' means create (overwrite if exists); 'w-' means create
                (fail if exists).
            path: path within the store to open.

        Example:
            >>> from pathlib import Path
            >>> ez_list = EZList.from_list([1, "some text", Path("/some/path.txt")], name="data")
            >>> ez_list.save("/tmp/list", overwrite=True)
            >>> EZList.open("/tmp/list", name="data")
            EZList[1, 'some text', PosixPath('/some/path.txt')]
        """
        path = f"{path.rstrip('/')}/{name}" if path else name
        arr = zarr.open_array(store, mode=mode, path=path)

        arr_type = cast(str, arr.attrs.get(Attribute.EZType, UNKNOWN))
        if arr_type != EZType.List:
            raise ValueError(f"Cannot read EZList from Array marked with type '{arr_type}', should be 'list'.")

        return cls(arr)

    @override
    def save(
        self,
        store: StoreLike,
        *,
        name: str | None = None,
        mode: AccessModeLiteral = "a",
        path: str | None = None,
        overwrite: bool = False,
    ) -> None:
        """
        Save this EZList to a local file system.

        Args:
            store: Store, path to a directory or name of a zip file.
            name: name for the EZList, to use inside the store. (default: same as name used during EZLits creation)
            mode: Persistence mode: 'r' means read only (must exist); 'r+' means
                read/write (must exist); 'a' means read/write (create if doesn't
                exist); 'w' means create (overwrite if exists); 'w-' means create
                (fail if exists).
            path: path within the store where the EZList will be saved.
            overwrite: overwrite EZList if a group with name `name` already exists ? (default: False)

        Example:
            >>> from pathlib import Path
            >>> ez_list = EZList.from_list([1, "some text", Path("/some/path.txt")], name="data")
            >>> ez_list.save("/tmp/list", overwrite=True)
            >>> Path("/tmp/list").exists()
            True
        """
        if name is None:
            name = self._array.name

        path = f"{path.rstrip('/')}/{name}" if path else name

        with warnings.catch_warnings(action="ignore", category=UnstableSpecificationWarning):
            arr = zarr.create_array(
                store=store,
                name=path,
                shape=(self._array.shape[0],),
                dtype=PyObject(),
                serializer=PyObjectCodec(),
                attributes={Attribute.EZType: EZType.List},
                overwrite=overwrite,
            )
        arr[()] = self._array[()]

    @override
    def copy(self) -> list[T]:
        """
        Convert this EZList into a Python list, loading all the data into memory.

        Example:
            >>> from pathlib import Path
            >>> ez_list = EZList.from_list([1, "some text", Path("/some/path.txt")], name="data")
            >>> ez_list.copy()
            [1, 'some text', PosixPath('/some/path.txt')]
        """
        return self._array[:].tolist()  # pyright: ignore[reportUnknownVariableType, reportAttributeAccessIssue]

    @classmethod
    def defer(cls, values: Collection[Any]) -> DeferredCreationFunc:
        r"""
        Create an EZList from a list-like object, but defer its storage.
        This function returns a function that can be called to finalize the EZList creation and storage,
        with parameters:
            grp: (zarr.Group) a group in which to store the EZList
            name: (str) name for the EZList, to use inside the store.
            overwrite: (bool) overwrite object if a group with name `name` already exists ? (default: False)

        This is usually used as a convenience function to create an EZList in an existing EZDict.

        Args:
            values: list-like collection of values to store in an EZList.

        Example:
            >>> from ezarr import EZDict
            >>> from pathlib import Path
            >>> ez_list = EZList.from_list([1, "some text", Path("/some/path.txt")], name="data")

            >>> d = EZDict(zarr.open_group({}))
            >>> d["list"] = EZList.defer([1, 2, 3])
            >>> repr(d)
            'EZDict{\n\tlist: EZList[1, 2, 3]\n}'
        """
        return deferred_EZList(values)

    @override
    def append(self, value: T) -> None:
        """
        Append a value to the end of an EZList.

        Args:
            value: value to be appended.

        Example:
            >>> from pathlib import Path
            >>> ez_list = EZList.from_list([1, "some text", Path("/some/path.txt")], name="data")

            >>> ez_list.append(3.14)
            >>> ez_list
            EZList[1, 'some text', PosixPath('/some/path.txt'), 3.14]

            >>> ez_list.append([1, 2, 3])
            >>> ez_list
            EZList[1, 'some text', PosixPath('/some/path.txt'), 3.14, array([1, 2, 3])]
        """
        if isinstance(value, list | SupportsArray):
            value = ListWrapper(value)  # pyright: ignore[reportAssignmentType, reportUnknownArgumentType]

        with warnings.catch_warnings(action="ignore", category=UnstableSpecificationWarning):
            self._array.append(np.asarray([value]))

    @override
    def extend(self, values: Iterable[T]) -> None:
        """
        Append multiple values to the end of an EZList.

        Args:
            values: values to be appended.

        Example:
            >>> from pathlib import Path
            >>> ez_list = EZList.from_list([1, "some text", Path("/some/path.txt")], name="data")

            >>> ez_list.extend([3.14, 2.71, [-1, -2]])
            >>> ez_list
            EZList[1, 'some text', PosixPath('/some/path.txt'), 3.14, 2.71, array([-1, -2])]
        """
        values_np = np.asarray([ListWrapper(v) if isinstance(v, list | SupportsArray) else v for v in values])  # pyright: ignore[, reportUnknownArgumentType]

        with warnings.catch_warnings(action="ignore", category=UnstableSpecificationWarning):
            self._array.append(values_np)

    @override
    def insert(self, index: int, value: T) -> None:
        """
        Insert a value in an EZList before `index`.

        Args:
            value: value to be inserted.

        Example:
            >>> from pathlib import Path
            >>> ez_list = EZList.from_list([1, "some text", Path("/some/path.txt")], name="data")

            >>> ez_list.insert(1, [1, 2, 3])
            >>> ez_list
            EZList[1, array([1, 2, 3]), 'some text', PosixPath('/some/path.txt')]
        """
        if index < 0:
            index = len(self) + index

        if isinstance(value, list | SupportsArray):
            value = ListWrapper(value)  # pyright: ignore[reportAssignmentType, reportUnknownArgumentType]

        with warnings.catch_warnings(action="ignore", category=UnstableSpecificationWarning):
            self._array.resize(self._array.shape[0] + 1)
        self._array[index + 1 :] = self._array[index:-1]

        self._array[index] = value  # pyright: ignore[reportArgumentType]



import numpy as np
import numpy.typing as npt
import zarr


def take_1d(
    arr: zarr.Array,
    indexer: npt.NDArray[np.intp],
    fill_value=None,
    allow_fill: bool = True,
    mask: npt.NDArray[np.bool_] | None = None,
) -> zarr.Array:
    """
    Specialized version for 1D arrays. Differences compared to `take_nd`:

    - Assumes input array has already been converted to numpy array / EA
    - Assumes indexer is already guaranteed to be intp dtype ndarray
    - Only works for 1D arrays

    To ensure the lowest possible overhead.

    Note: similarly to `take_nd`, this function assumes that the indexer is
    a valid(ated) indexer with no out of bound indices.

    Parameters
    ----------
    arr : np.ndarray or ExtensionArray
        Input array.
    indexer : ndarray
        1-D array of indices to take (validated indices, intp dtype).
    fill_value : any, default np.nan
        Fill value to replace -1 values with
    allow_fill : bool, default True
        If False, indexer is assumed to contain no -1 values so no filling
        will be done.  This short-circuits computation of a mask. Result is
        undefined if allow_fill == False and -1 is present in indexer.
    mask : np.ndarray, optional, default None
        If `allow_fill` is True, and the mask (where indexer == -1) is already
        known, it can be passed to avoid recomputation.
    """
    if fill_value is None:
        fill_value = np.nan

    if allow_fill:
        assert arr.ndim == 1
        assert mask is not None
        if len(arr) < len(mask):
            np.insert(arr, np.where(mask), fill_value)

    return arr[indexer]



import zarr
import numpy as np

import ezarr.inplace as ezi


def test_delete():
    array = zarr.create_array({}, data=np.array([[1, 2, 3], [4, 5, 6]]))
    assert np.array_equal(ezi.delete(array, 1, axis=1), np.array([[1, 3], [4, 6]]))



from __future__ import annotations

from numbers import Number
import pickle
from collections.abc import Collection, Mapping
from typing import Any, cast
import warnings

import numpy as np
import zarr
import zarr.api.asynchronous
from zarr.errors import UnstableSpecificationWarning

import ezarr
from ezarr.dict import DictParameters
from ezarr.names import UNKNOWN, Attribute, EZType
from ezarr.types import DeferredCreationFunc, PyObject, PyObjectCodec, SupportsEZRead, SupportsEZWrite


def read_object(grp: zarr.Group, *, name: str, path: str = "") -> Any:
    path = f"{path.lstrip('/')}/{name}" if path else name
    data = grp[path]
    ez_type = cast(str, data.attrs.get(Attribute.EZType, UNKNOWN))

    if isinstance(data, zarr.Group):
        if ez_type != EZType.Object:
            return ezarr.EZDict[Any](data)

        ez_class = cast(zarr.Array | None, data.get(Attribute.EZClass))
        if ez_class is None:
            raise ValueError("Cannot read object with unknown class.")

        assert isinstance(ez_class, zarr.Array)
        data_class = pickle.loads(ez_class[()])  # pyright: ignore[reportArgumentType]

        if not issubclass(data_class, SupportsEZRead):
            raise ValueError(f"Can't read {data_class} since it does not implement the '__ez_read__' method.")

        return data_class.__ez_read__(ezarr.EZDict(data))

    if data.ndim == 0:
        return data[()]

    if ez_type == EZType.List:
        return ezarr.EZList[Any](data)

    return data


def write_object[T](
    grp: zarr.Group | ezarr.EZDict[Any],
    *,
    obj: Any,
    name: str,
    path: str = "",
    overwrite: bool = False,
    parameters: DictParameters | None = None,
) -> None:
    """
    Save any object in a zarr.Group

    Args:
        grp: a zarr.Group.
        obj: a Python object to be saved.
        name: name for the object, to use inside the Group.
        path: [optional] path within the Group where the object will be saved.
        overwrite: overwrite if a group with name `name` already exists ? (default: False)
        parameters: optional parameters passed during array creation.
    """
    grp = grp.group if isinstance(grp, ezarr.EZDict) else grp

    if path:
        grp = grp.require_group(path, overwrite=overwrite)

    match obj:
        case SupportsEZWrite():
            subgroup = grp.create_group(name, overwrite=overwrite)
            obj.__ez_write__(ezarr.EZDict(subgroup))
            subgroup.attrs[Attribute.EZType] = EZType.Object

            klass = pickle.dumps(type(obj), protocol=pickle.HIGHEST_PROTOCOL)

            subgroup.create_array(Attribute.EZClass, data=np.void(klass), overwrite=True)  # pyright: ignore[reportArgumentType]

        case DeferredCreationFunc():
            obj(grp, name, overwrite)

        case Mapping():
            subgroup = grp.create_group(name, overwrite=overwrite)
            write_objects(subgroup, "", parameters=parameters, **obj)

        case zarr.Array() | np.ndarray():
            full_name = f"{grp.path.rstrip('/')}/{name}" if grp.path else f"{name}"
            grp._sync(zarr.api.asynchronous.from_array(grp.store, name=full_name, data=obj))  # pyright: ignore[reportPrivateUsage]

        case Collection() | Number():
            with warnings.catch_warnings(action="ignore", category=UnstableSpecificationWarning):
                grp.create_array(name, data=np.array(obj), overwrite=overwrite, **(parameters or {}))

        case _:
            with warnings.catch_warnings(action="ignore", category=UnstableSpecificationWarning):
                arr = grp.create_array(
                    name, shape=(), dtype=PyObject(), serializer=PyObjectCodec(), overwrite=overwrite
                )
            arr[()] = obj


def write_objects[T](
    grp: zarr.Group | ezarr.EZDict[Any], path: str = "", parameters: DictParameters | None = None, **kwargs: Any
) -> None:
    for name, value in kwargs.items():
        write_object(grp, obj=value, name=name, path=path, parameters=parameters)



"""
Functions for operating on zarr.Arrays inplace
"""

from typing import Any, cast

import numpy as np
import numpy.typing as npt
import zarr
from numpy._typing import _ArrayLikeInt_co  # pyright: ignore[reportPrivateUsage]


def delete[A: zarr.Array | npt.NDArray[Any], T: np.generic](
    arr: A, obj: _ArrayLikeInt_co | slice, axis: int | None = None
) -> A:
    if not isinstance(arr, zarr.Array):
        return np.delete(arr, obj, axis)

    if arr.ndim > 1 and axis is None:
        raise NotImplementedError("Cannot flatten zarr.Array while working inplace")

    if axis is None:
        axis = 0

    if not isinstance(obj, int | np.integer):
        raise NotImplementedError

    obj = cast(np.integer, obj)

    if obj > arr.shape[axis]:
        raise IndexError(f"Index {obj} is out of bounds for axis {axis} with size {arr.shape[axis]}.")

    elif obj < 0:
        obj = cast(np.integer, obj + arr.shape[axis])

    prefix = (slice(None),) * axis
    if obj < (arr.shape[axis] - 1):
        # transfer data one row to the left, starting from the column after the one to delete
        # matrix | 0 1 2 3 4 | with index of the column to delete = 2
        #   ==>  | 0 1 3 4 . |
        index_dest = prefix + (slice(obj, -1),)
        index_source = prefix + (slice(obj + 1, None),)
        arr[index_dest] = arr[index_source]

    # resize the arrays to drop the extra column at the end
    # matrix | 0 1 3 4 . |
    #   ==>  | 0 1 3 4 |
    new_shape = arr.shape[:axis] + (arr.shape[axis] - 1,) + arr.shape[axis + 1 :]
    arr.resize(new_shape)

    return arr



from typing import Any, Self, cast, override
import numpy as np
import zarr

import ezarr
from ezarr.types import SupportsEZReadWrite


def test_dict_creation(ez_dict: ezarr.EZDict[Any]):
    assert isinstance(ez_dict, ezarr.EZDict)


def test_dict_can_iterate_through_keys(ez_dict: ezarr.EZDict[Any]):
    assert sorted(iter(ez_dict)) == ["a", "b", "c", "f"]


def test_dict_has_correct_keys(ez_dict: ezarr.EZDict[Any]):
    assert sorted(ez_dict.keys()) == ["a", "b", "c", "f"]


def test_dict_can_get_regular_values(ez_dict: ezarr.EZDict[Any]):
    assert ez_dict["a"] == 1


def test_dict_can_get_string(ez_dict: ezarr.EZDict[Any]):
    assert ez_dict["c"]["d"] == "test"


def test_dict_can_get_Array(ez_dict: ezarr.EZDict[Any]):
    assert isinstance(ez_dict["c"]["e"], zarr.Array)


def test_dict_should_return_string(ez_dict: ezarr.EZDict[Any]):
    assert isinstance(ez_dict["c"]["d"], str)


def test_dict_gets_nested_ez_dicts(ez_dict: ezarr.EZDict[Any]):
    assert isinstance(ez_dict["c"], ezarr.EZDict)


def test_dict_has_correct_values(ez_dict: ezarr.EZDict[Any]):
    assert ez_dict["a"] == 1
    assert np.array_equal(ez_dict["b"], [1, 2, 3])
    assert sorted(ez_dict["c"].keys()) == ["d", "e"]
    assert ez_dict["f"].shape == (10, 10, 10)


def test_dict_can_set_regular_value(ez_dict: ezarr.EZDict[Any]):
    ez_dict["a"] = 5

    assert ez_dict["a"] == 5


def test_dict_can_set_array_value(ez_dict: ezarr.EZDict[Any]):
    ez_dict["b"][1] = 6

    assert np.array_equal(ez_dict["b"], [1, 6, 3])


def test_dict_can_set_new_regular_value(ez_dict: ezarr.EZDict[Any]):
    ez_dict["x"] = 9

    assert ez_dict["x"] == 9


def test_dict_can_set_new_array(ez_dict: ezarr.EZDict[Any]):
    ez_dict["y"] = np.array([1, 2, 3])

    assert np.array_equal(ez_dict["y"], [1, 2, 3])


def test_dict_can_set_new_zarray(ez_dict: ezarr.EZDict[Any]):
    arr = zarr.create_array({}, data=np.array([1, 2, 3]))
    ez_dict["zy"] = arr

    assert np.array_equal(ez_dict["zy"], [1, 2, 3])


def test_dict_can_set_new_dict(ez_dict: ezarr.EZDict[Any]):
    ez_dict["z"] = {"l": 10, "m": [10, 11, 12], "n": {"o": 13}}

    assert (
        isinstance(ez_dict["z"], ezarr.EZDict)
        and ez_dict["z"]["l"] == 10
        and np.array_equal(ez_dict["z"]["m"], [10, 11, 12])
        and isinstance(ez_dict["z"]["n"], ezarr.EZDict)
        and ez_dict["z"]["n"]["o"] == 13
    )


def test_dict_can_replace_dict(ez_dict: ezarr.EZDict[Any]):
    ez_dict["c"] = {"d2": "test", "e": np.arange(10, 20)}

    assert isinstance(ez_dict["c"], ezarr.EZDict)
    assert "d" not in ez_dict["c"].keys()
    assert ez_dict["c"]["d2"] == "test"
    assert np.array_equal(ez_dict["c"]["e"], np.arange(10, 20))


def test_dict_can_union(ez_dict: ezarr.EZDict[Any]):
    ez_dict["c"] |= {"d": "new_val", "g": -1, "h": {"i": "nested"}}

    assert sorted(ez_dict["c"].keys()) == ["d", "e", "g", "h"]
    assert ez_dict["c"]["d"] == "new_val"
    assert np.array_equal(ez_dict["c"]["e"], np.arange(100))
    assert ez_dict["c"]["g"] == -1
    assert ez_dict["c"]["h"]["i"] == "nested"


def test_dict_can_delete_regular_value(ez_dict: ezarr.EZDict[Any]):
    del ez_dict["a"]

    assert "a" not in ez_dict.keys()


def test_dict_can_delete_array(ez_dict: ezarr.EZDict[Any]):
    del ez_dict["b"]

    assert "b" not in ez_dict.keys()


def test_dict_can_delete_dict(ez_dict: ezarr.EZDict[Any]):
    del ez_dict["c"]

    assert "c" not in ez_dict.keys()


def test_dict_copy_should_be_regular_dict(ez_dict: ezarr.EZDict[Any]):
    c = ez_dict.copy()

    assert isinstance(c, dict)


def test_dict_copy_should_have_same_keys(ez_dict: ezarr.EZDict[Any]):
    c = ez_dict.copy()

    assert c.keys() == ez_dict.keys()


def test_dict_copy_nested_ez_dict_should_be_dict(ez_dict: ezarr.EZDict[Any]):
    c = ez_dict.copy()

    assert isinstance(c["c"], dict)


def test_dict_copy_dataset_proxy_should_be_array(ez_dict: ezarr.EZDict[Any]):
    c = ez_dict.copy()

    assert type(c["b"]) == np.ndarray  # noqa: E721


class ComplexObject(SupportsEZReadWrite):
    def __init__(self, value: int):
        self.value: int = value

    @override
    def __repr__(self) -> str:
        return f"CO({self.value})"

    @override
    def __ez_write__(self, values: ezarr.EZDict[Any]) -> None:
        values.attrs["value"] = self.value

    @override
    @classmethod
    def __ez_read__(cls, values: ezarr.EZDict[Any]) -> Self:
        return cls(cast(int, values.attrs["value"]))


def test_dict_can_store_complex_objects(ez_dict: ezarr.EZDict[Any]):
    ez_dict["g"] = {"a": ComplexObject(1), "b": ComplexObject(2)}

    assert isinstance(ez_dict["g"]["a"], ComplexObject)


def test_dict_can_be_destructured_in_pattern_matching(ez_dict: ezarr.EZDict[Any]):
    match ez_dict:
        case ezarr.EZDict(grp):  # pyright: ignore[reportUnknownVariableType, reportGeneralTypeIssues]
            assert isinstance(grp, zarr.Group)

        case _:
            raise ValueError



import numpy as np

from ezarr.dataframe import EZDataFrame


def test_can_map_function(ezdf: EZDataFrame) -> None:
    res = ezdf.col_int.map(lambda x: x**2)

    assert np.array_equal(res, [1, 4, 9])



from __future__ import annotations

import re
import warnings
from collections.abc import Callable, Generator, Hashable, Iterable
from typing import Any, Self, override

import numpy as np
import numpy.typing as npt
import pandas as pd
import zarr
import zarr.abc.store
from zarr.core.dtype import VariableLengthUTF8
import zarr.storage
from pandas.core.generic import NDFrame
from pandas.core.internals.array_manager import BaseArrayManager
from zarr.codecs import VLenUTF8Codec
from zarr.core.common import AccessModeLiteral
from zarr.errors import UnstableSpecificationWarning
from zarr.storage import StoreLike  # pyright: ignore[reportUnknownVariableType]

import ezarr
from ezarr.dataframe.manager import EZArrayManager
from ezarr.types import SupportsEZReadWrite


def _get_arrays(store: zarr.Group, arrays: Iterable[tuple[Hashable, Any]]) -> Generator[zarr.Array, None, None]:
    for name, arr in arrays:
        array = np.asarray(arr)

        if array.dtype == object:
            if array.size:
                array = array.astype(type(array.flat[0]))  # pyright: ignore[reportUnknownArgumentType, reportUnknownVariableType]

            if array.dtype == object:
                raise TypeError(
                    f'Failed interpreting the "Object" data type to a non ambiguous data type for array {array}'
                )

        if np.issubdtype(array.dtype, np.str_):  # pyright: ignore[reportUnknownArgumentType]
            z_arr = store.create_array(
                str(name), shape=array.shape, dtype=VariableLengthUTF8(), serializer=VLenUTF8Codec()
            )
            z_arr[:] = array
            yield z_arr

        else:
            yield store.create_array(str(name), data=array)


class EZDataFrame(pd.DataFrame, SupportsEZReadWrite):  # EZObject[pd.DataFrame],
    """
    Proxy for pandas DataFrames for storing data in zarr stores.
    """

    _internal_names: list[str] = ["_data_store"] + pd.DataFrame._internal_names
    _internal_names_set: set[str] = {"_data_store"} | pd.DataFrame._internal_names_set

    _data_store: zarr.Group

    @property
    def _constructor(self) -> Callable[..., EZDataFrame]:
        def inner(df: Any) -> EZDataFrame:
            if not isinstance(df, EZDataFrame):
                raise ValueError("EZDataFrame constructor not properly called")

            return df

        return inner

    def _constructor_from_mgr(self, mgr: BaseArrayManager, axes: list[pd.Index]) -> pd.DataFrame:
        if not isinstance(mgr, EZArrayManager):
            df: pd.DataFrame = pd.DataFrame._from_mgr(mgr, axes=axes)  # pyright: ignore[reportAttributeAccessIssue]

        else:
            df = EZDataFrame._from_mgr(mgr, axes=axes)  # pyright: ignore[reportAttributeAccessIssue]

        if isinstance(self, pd.DataFrame):
            # This would also work `if self._constructor is DataFrame`, but
            #  this check is slightly faster, benefiting the most-common case.
            return df

        elif type(self).__name__ == "GeoDataFrame":
            # Shim until geopandas can override their _constructor_from_mgr
            #  bc they have different behavior for Managers than for DataFrames
            return self._constructor(mgr)

        # We assume that the subclass __init__ knows how to handle a
        #  pd.DataFrame object.
        return self._constructor(df)

    def _constructor_sliced_from_mgr(self, mgr: BaseArrayManager, axes: list[pd.Index[Any]]) -> pd.Series[Any]:
        ser: pd.Series = pd.Series._from_mgr(mgr, axes)
        ser._name = None  # caller is responsible for setting real name

        if isinstance(self, pd.DataFrame):
            return ser

        return self._constructor_sliced(ser)

    def __init__(
        self,
        data: pd.DataFrame
        | dict[Hashable, npt.NDArray[Any]]
        | ezarr.EZDict[Any]
        | npt.NDArray[Any]
        | zarr.Array
        | tuple[EZArrayManager, zarr.Group]
        | None = None,
        index: zarr.Array | pd.Index | None = None,
        columns: zarr.Array | pd.Index | None = None,
    ):
        match data:
            case ezarr.EZDict():
                assert index is None, (
                    "Cannot provide the `index` parameter when also providing the `data` parameter as EZDict."
                )
                assert columns is None, (
                    "Cannot provide the `columns` parameter when also providing the `data` parameter as EZDict."
                )

                _index: pd.Index = pd.Index(data["index"].copy())
                _columns: pd.Index = pd.Index(data["arrays"].keys())
                arrays: list[zarr.Array] = [arr for arr in data["arrays"].values()]

                store = data.group
                mgr = EZArrayManager(arrays, [_index, _columns], group=store)

            case pd.DataFrame():
                _index = data.index if index is None else pd.Index(np.array(index))
                _columns = data.columns if columns is None else pd.Index(np.array(columns))

                store = zarr.create_group({})
                with warnings.catch_warnings(action="ignore", category=UnstableSpecificationWarning):
                    arrays = list(_get_arrays(store, data.to_dict(orient="series").items()))

                mgr = EZArrayManager(arrays, [_index, _columns], group=store)

            case np.ndarray() | zarr.Array():
                _index = pd.RangeIndex(start=0, stop=data.shape[0]) if index is None else pd.Index(np.array(index))
                _columns = (
                    pd.RangeIndex(start=0, stop=data.shape[1]) if columns is None else pd.Index(np.array(columns))
                )

                store = zarr.create_group({})
                with warnings.catch_warnings(action="ignore", category=UnstableSpecificationWarning):
                    arrays = list(_get_arrays(store, zip(_columns, np.transpose(np.asarray(data)))))

                mgr = EZArrayManager(arrays, [_index, _columns], group=store)

            case dict():
                assert columns is None, (
                    "Cannot provide the `columns` parameter when also providing the `data` parameter as dict."
                )

                _columns = pd.Index(data.keys())

                store = zarr.create_group({})
                arrays = list(_get_arrays(store, data.items()))

                _index = pd.RangeIndex(start=0, stop=arrays[0].shape[0]) if index is None else pd.Index(np.array(index))

                mgr = EZArrayManager(arrays, [_index, _columns], group=store)

            case None:
                _index = pd.RangeIndex(0) if index is None else pd.Index(np.array(index))
                _columns = pd.RangeIndex(0) if columns is None else pd.Index(np.array(columns))
                arrays = []

                store = zarr.create_group({})
                mgr = EZArrayManager(arrays, [_index, _columns], group=store)

            case (EZArrayManager() as mgr, zarr.Group() as store):
                pass

            case _:  # pyright: ignore[reportUnnecessaryComparison]
                raise TypeError(f"Invalid type '{type(data)}' for 'data' argument.")  # pyright: ignore[reportUnreachable]

        object.__setattr__(self, "_data_store", store)
        NDFrame.__init__(self, mgr)  # pyright: ignore[reportCallIssue]

    @override
    def __finalize__(self, other: EZDataFrame, method: str | None = None, **kwargs: Any) -> pd.DataFrame:  # pyright: ignore[reportIncompatibleMethodOverride]
        super().__finalize__(other, method, **kwargs)

        if method == "copy":
            return other

        return self

    @staticmethod
    def _get_mode(store: zarr.abc.store.Store) -> str:
        match store:
            case zarr.storage.FsspecStore():
                return "FSSpec"
            case zarr.storage.GpuMemoryStore():
                return "GPU"
            case zarr.storage.LocalStore():
                return "Local"
            case zarr.storage.LoggingStore(store=inner_store):  # pyright: ignore[reportUnknownVariableType]
                return f"Logging:{EZDataFrame._get_mode(inner_store)}"  # pyright: ignore[reportUnknownArgumentType]
            case zarr.storage.MemoryStore():
                return "RAM"
            case zarr.storage.ObjectStore():
                return "Object"
            case zarr.storage.WrapperStore(store=inner_store):  # pyright: ignore[reportUnknownVariableType]
                return f"Wrapper:{EZDataFrame._get_mode(inner_store)}"  # pyright: ignore[reportUnknownArgumentType]
            case zarr.storage.ZipStore():
                return "Zip"
            case _:
                return type(store).__name__

    @override
    def __repr__(self) -> str:
        if self.empty:
            return f"""\
Empty {type(self).__name__}
Columns: {self.columns.tolist()}
Index: {self.index.tolist()}
[{self._get_mode(self._data_store.store)}]"""

        repr_ = repr(self.iloc[:5].copy())
        re.sub(r"\n\n\[.*\]$", "", repr_)
        return (
            repr_
            + ("\n..." if self.shape[1] > 5 else "")
            + f"\n[{self._get_mode(self._data_store.store)}]\n[{len(self.index)} rows x {len(self.columns)} columns]"
        )

    @override
    def __setitem__(self, key, value) -> None:
        return super().__setitem__(key, value)

    @override
    def __ez_write__(self, values: ezarr.EZDict[Any]) -> None:
        if values.group.store == self._data_store.store:
            return

        _index = self.index.values
        if _index.dtype == object:
            _index = _index.astype(str)

        values["index"] = _index
        values["arrays"] = {str(k): v.values for k, v in self.to_dict(orient="series").items()}
        values["arrays"].attrs["columns_order"] = self.columns.to_list()

    @classmethod
    @override
    def __ez_read__(cls, values: ezarr.EZDict[Any]) -> Self:
        arrays = values["arrays"]
        columns_order = values["arrays"].attrs["columns_order"]

        mgr = EZArrayManager(
            [arr for (_, arr) in sorted(arrays.group.arrays(), key=lambda name_arr: columns_order.index(name_arr[0]))],
            [pd.Index(values["index"][:]), pd.Index(columns_order)],
            group=values.group,
        )

        return cls((mgr, values.group))

    @property
    def data(self) -> ezarr.EZDict[Any]:
        return ezarr.EZDict(self._data_store)

    @classmethod
    def open(
        cls, store: StoreLike | None = None, *, name: str, mode: AccessModeLiteral = "a", path: str | None = None
    ) -> Self:
        """
        Open this object from a store.

        Args:
            store: Store, path to a directory or name of a zip file.
            name: name for the object, to use inside the store.
            mode: Persistence mode: 'r' means read only (must exist); 'r+' means
                read/write (must exist); 'a' means read/write (create if doesn't
                exist); 'w' means create (overwrite if exists); 'w-' means create
                (fail if exists).
            path: path within the store to open.
        """
        path = f"{path.rstrip('/')}/{name}" if path else name
        return cls.__ez_read__(ezarr.EZDict(zarr.open_group(store, mode=mode, path=path)))

    def save(
        self,
        store: StoreLike,
        *,
        name: str,
        mode: AccessModeLiteral = "a",
        path: str | None = None,
        overwrite: bool = False,
    ) -> None:
        """
        Save this object to a local file system.

        Args:
            store: Store, path to a directory or name of a zip file.
            name: name for the object, to use inside the store.
            mode: Persistence mode: 'r' means read only (must exist); 'r+' means
                read/write (must exist); 'a' means read/write (create if doesn't
                exist); 'w' means create (overwrite if exists); 'w-' means create
                (fail if exists).
            path: path within the store where the object will be saved.
            overwrite: overwrite object if a group with name `name` already exists ? (default: False)
        """
        path = f"{path.rstrip('/')}/{name}" if path else name

        if not overwrite:
            mode = "w-"

        self.__ez_write__(ezarr.EZDict(zarr.open_group(store, mode=mode, path=path)))



import importlib

from ezarr import inplace, io
from ezarr.dict import EZDict
from ezarr.list import EZList
from ezarr.types import PyObject, PyObjectCodec, SupportsEZRead, SupportsEZReadWrite, SupportsEZWrite

importlib.import_module("ezarr.patch")

__all__ = [
    "inplace",
    "EZDict",
    "EZList",
    "io",
    "PyObject",
    "PyObjectCodec",
    "SupportsEZRead",
    "SupportsEZReadWrite",
    "SupportsEZWrite",
]



import pytest
import numpy as np
import pandas as pd
from ezarr.dataframe import EZDataFrame


def test_loc(ezdf: EZDataFrame) -> None:
    assert np.all(ezdf.loc[2].values == np.array([3, "def", 3.5], dtype=object))


def test_loc_multi(ezdf: EZDataFrame) -> None:
    assert np.all(ezdf.loc[[2, 1]].values == np.array([[3, "def", 3.5], [2, "bc", 2.5]], dtype=object))


def test_loc_single_value(ezdf: EZDataFrame) -> None:
    value = ezdf.loc[1, "col_str"]
    assert value.ndim == 0  # pyright: ignore[reportAttributeAccessIssue]
    assert np.issubdtype(value.dtype, np.dtypes.StringDType())  # pyright: ignore[reportUnknownArgumentType, reportAttributeAccessIssue]
    assert value == "bc"


def test_get_column(ezdf: EZDataFrame) -> None:
    assert ezdf["col_str"].equals(pd.Series(["a", "bc", "def"], name="col_str"))


def test_get_column_regular_df(ezdf: EZDataFrame):
    df = pd.DataFrame({"col_str": ["a", "bc", "def"]})
    assert df["col_str"].equals(pd.Series(["a", "bc", "def"], name="col_str"))

    assert ezdf["col_str"].equals(df["col_str"])


@pytest.mark.xfail
def test_compare_with_regular_df(ezdf: EZDataFrame):
    # TODO: maybe one day find a fix, for now can only use h5dataframe.equals(), pd.DataFrame.equals() fails
    df = pd.DataFrame({"col_str": ["a", "bc", "def"]})
    assert df["col_str"].equals(ezdf["col_str"])


def test_iterrows(ezdf: EZDataFrame):
    _, row = next(ezdf.iterrows())
    assert row.equals(pd.Series([1, "a", 1.5], index=["col_int", "col_str", "col_float"]))



import importlib
from sys import stderr
import sys

try:
    importlib.import_module("pandas")

except ModuleNotFoundError:
    print("Please install ezarr[pandas] to use dataframes", file=stderr)
    sys.exit(1)


from ezarr.dataframe.dataframe import EZDataFrame

importlib.import_module("ezarr.dataframe.patch")


__all__ = ["EZDataFrame"]



from typing import Any, Literal
from collections.abc import Callable

import numpy as np
import numpy.typing as npt
import zarr


def zarr_astype(self: zarr.Array, dtype: npt.DTypeLike, order: Literal["K", "A", "C", "F"] = "K", copy: bool = True):
    return self[:].astype(dtype, order=order, copy=copy)  # pyright: ignore[reportUnknownVariableType]


# pyright: reportAttributeAccessIssue=false, reportUnknownLambdaType=false
zarr.Array.__len__ = lambda self: self.shape[0]
zarr.Array.astype = zarr_astype
zarr.Array.copy = lambda self: self[:]


def _inplace(array: zarr.Array, func: Callable[..., Any], value: Any) -> zarr.Array:
    if array.read_only:
        raise ValueError("assignment destination is read-only")

    for chunk in array._iter_chunk_regions():  # pyright: ignore[reportPrivateUsage]
        array[chunk] = func(array[chunk], value)

    return array


# pyright: reportUnknownArgumentType=false
zarr.Array.__add__ = lambda self, other: np.add(self, other)
zarr.Array.__iadd__ = lambda self, other: _inplace(self, np.add, other)
zarr.Array.__sub__ = lambda self, other: np.subtract(self, other)
zarr.Array.__isub__ = lambda self, other: _inplace(self, np.subtract, other)
zarr.Array.__mul__ = lambda self, other: np.multiply(self, other)
zarr.Array.__imul__ = lambda self, other: _inplace(self, np.multiply, other)
zarr.Array.__truediv__ = lambda self, other: np.divide(self, other)
zarr.Array.__itruediv__ = lambda self, other: _inplace(self, np.divide, other)
zarr.Array.__mod__ = lambda self, other: np.mod(self, other)
zarr.Array.__imod__ = lambda self, other: _inplace(self, np.mod, other)
zarr.Array.__pow__ = lambda self, other: np.pow(self, other)
zarr.Array.__ipow__ = lambda self, other: _inplace(self, np.power, other)
zarr.Array.__or__ = lambda self, other: np.logical_or(self, other)
zarr.Array.__ior__ = lambda self, other: _inplace(self, np.logical_or, other)
zarr.Array.__and__ = lambda self, other: np.logical_and(self, other)
zarr.Array.__iand__ = lambda self, other: _inplace(self, np.logical_and, other)
zarr.Array.__xor__ = lambda self, other: np.logical_xor(self, other)
zarr.Array.__ixor__ = lambda self, other: _inplace(self, np.logical_xor, other)
